"""Code for generating an HTML report file with a table containing links to
Proksee projects and images for each sample. A single genome viewer is
positioned to the right of the table.

The output directory will be structured as in the following example:

    output_directory/
        cgview-js_code/
            ...
        data/
            genome_name_1.js
            genome_name_2.js
            ...
        report.html
"""

import os
from typing import Any
from typing import Dict


def generate_report_html(output_dir: str, genome_info: Dict[str, Any]) -> None:
    """
    Generates an HTML report file with a table containing links to Proksee
    projects and images for each sample. A single genome viewer is positioned
    to the right of the table.
    """
    assert os.path.isdir(output_dir), f"Output directory does not exist: {output_dir}"
    output_file = os.path.join(output_dir, "report.html")

    with open(output_file, "w") as file:
        # Write the DOCTYPE, html, head sections with CSS, and JavaScript imports
        file.write(
            """<!DOCTYPE html>
<html lang="en">
  <head>

    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">

    <!-- Import CGView.js code -->
    <link href="./cgview-js_code/docs/styles/bootstrap.min.css" rel="stylesheet">
    <script src='./cgview-js_code/docs/scripts/marked.min.js'></script>
    <script src='./cgview-js_code/docs/scripts/general.js'></script>
    <script src="./cgview-js_code/docs/scripts/d3.min.js"></script>
    <script src='./cgview-js_code/docs/dist/cgview.min.js'></script>
    <link rel="stylesheet" href="./cgview-js_code/docs/dist/cgview.css" />
    <link rel="stylesheet" href="./cgview-js_code/docs/styles/controls.css" />

    <title>Proksee Batch</title>

    <style>
      header {
        background-color: #f8f9fa;
        padding: 10px;
        text-align: left;
      }

      header h1 {
        margin: 0;
        color: #333;
      }

      body, html {
        font-family: 'Helvetica', 'Arial', sans-serif;
        margin: 0;
        padding: 0;
        box-sizing: border-box;
      }

      .container {
        display: flex;
        margin: 0;
        width: 100%;
      }

      .side-table {
        margin: 0;
        flex-shrink: 0;
        width: 60%;
        max-width: none;
      }

      .scrollable-table {
        height: 90%;
        overflow-y: auto;
        margin-top: 10px;
      }

      table tr:nth-child(even) {
        background-color: #f2f2f2;
      }

      #sortable-table tr td:last-child {
          display: flex;
          justify-content: center;
          align-items: center;
          height: 100%;
      }

      .viewer-container {
        display: flex;
        flex-direction: column;
        align-items: center;
        width: 40%;
      }

      #my-viewer {
        width: 100%;
      }

      .cgv-controls {
        width: 100%;
        display: flex;
        justify-content: center;
      }

      .generate-link {
          color: blue;
          text-decoration: underline;
          cursor: pointer;
      }

      .generated-link {
        color: blue;
        text-decoration: underline;
        cursor: pointer;
      }

      .btn-toggle-legend,
      #btn-toggle-legend {
        content: url("data:image/svg+xml,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 30 30'%3E%3Ccircle cx='5' cy='7' r='2' fill='black'/%3E%3Cpath d='M10 7 h15' stroke='black' stroke-width='2'/%3E%3Ccircle cx='5' cy='15' r='2' fill='black'/%3E%3Cpath d='M10 15 h15' stroke='black' stroke-width='2'/%3E%3Ccircle cx='5' cy='23' r='2' fill='black'/%3E%3Cpath d='M10 23 h15' stroke='black' stroke-width='2'/%3E%3C/svg%3E");
      }

    </style>
  </head>
<body>
    <header>
        <h1>Proksee Batch</h1>
    </header>

    <main>
        <div class="container">
        <div class="side-table">
            <div class="scrollable-table">
            <table id="sortable-table">
                <tr>
                    <th onclick="sortTable(0)">Sample ID</th>
                    <th onclick="sortTable(1)">Total size (bp)</th>
                    <th onclick="sortTable(2)">Contigs</th>
                    <th onclick="sortTable(3)">GC content</th>
                    <th>Action</th>
                </tr>
        """
        )

        # Create table rows based on the js and svg files
        for genome_code_name, info in genome_info.items():
            genome_name = info["Name"]
            # description = info["Description"]
            total_size = str(info["Total size"])
            number_of_contigs = str(info["Number of contigs"])
            gc_content = str(info["GC content"])

            file.write(
                f"""
                    <tr id='data/{genome_code_name}.js'>
                        <td>{genome_name}</td>
                        <td>{total_size}</td>
                        <td>{number_of_contigs}</td>
                        <td>{gc_content}</td>
                        <td class="generate-link" onclick="generateProkseeLink(this, 'data/{genome_code_name}')">Generate Proksee Project</td>
                    </tr>
            """
            )

        # Close the table.
        file.write(
            """
                </table>
            </div>
        </div>
        """
        )

        # Add the genome viewer.
        file.write(
            """
        <!-- Genome Viewer and Controls Container -->
        <div class="viewer-container">
            <!-- Genome Viewer -->
            <div id='my-viewer'></div>
            <!-- Controls -->
            <div class='cgv-controls'>
            <div class='cgv-btn' id='btn-reset' title='Reset Map'></div>
            <div class='cgv-btn' id='btn-zoom-in' title='Zoom In'></div>
            <div class='cgv-btn' id='btn-zoom-out' title='Zoom Out'></div>
            <div class='cgv-btn' id='btn-move-left' title='Move Left/Counterclockwise'></div>
            <div class='cgv-btn' id='btn-move-right' title='Move Right/Clockwise'></div>
            <div class='cgv-btn' id='btn-toggle-format' title='Toggle Linear/Circular Format'></div>
            <div class='cgv-btn' id='btn-invert-colors' title='Invert Map Colors'></div>
            <div class='cgv-btn' id='btn-download' title='Download Map PNG'></div>
            <div class='cgv-btn' id='btn-toggle-labels' title='Toggle Labels'></div>
            <div class='cgv-btn' id='btn-toggle-legend' title='Toggle Legend'></div>
            </div>
        </div>
        """
        )

        # Close the main content.
        file.write(
            """
    </div>
  </main>
        """
        )

        # Add script content.
        file.write(
            """
    <script src="./cgview-js_code/docs/scripts/bootstrap.min.js"></script>
    <script src="./cgview-js_code/docs/scripts/controls.js"></script>
    <script>
        function autoResizeMyViewer() {
            const mainPadding = 5;
            function myResize() {
                const myViewer = document.querySelector('#my-viewer');
                const main = document.getElementsByTagName('main')[0];
                const mainWidth = main.clientWidth * 0.50 - mainPadding;

                const mainHeight = main.clientHeight - 50;

                const minHeight = 500; // Set a minimum height if necessary
                const height = Math.max(mainHeight, minHeight);

                cgv.resize(mainWidth, height);
            }

            window.onresize = myResize;
            window.onload = function () {
                setTimeout( () => {
                    myResize();
                }, 100);
            }
        }

        function createViewer() {
            // Create Viewer in default div: #my-viewer
            const cgv = new CGV.Viewer('#my-viewer', {height: 500});
            autoResizeMyViewer();
            window.cgv = cgv;
        }

        function loadDataFile(filePath) {
            var script = document.createElement('script');
            script.src = filePath;
            script.onload = function() {
            cgv.io.loadJSON(window.json);
            cgv.draw();
            };
            document.head.appendChild(script);
        }

        document.addEventListener('DOMContentLoaded', (event) => {
            createViewer();

            // Add click event listener to each table row
            const rows = document.querySelectorAll('.side-table table tr');
            rows.forEach(row => {
            row.addEventListener('click', () => {
                const filePath = row.id;
                loadDataFile(filePath);
            });
            });

            // Check if there is at least one row and load the first genome by default
            if (rows.length > 0) {
            const firstRowFilePath = rows[1].id;
            loadDataFile(firstRowFilePath);
            }
        });

        function sortTable(column) {
            var table, rows, switching, i, x, y, shouldSwitch, dir, switchcount = 0;
            table = document.getElementById("sortable-table");
            switching = true;
            // Set the sorting direction to ascending:
            dir = "asc";
            // Make a loop that will continue until no switching has been done:
            while (switching) {
            // Start by saying: no switching is done:
            switching = false;
            rows = table.rows;
            // Loop through all table rows (except the first, which contains table headers):
            for (i = 1; i < (rows.length - 1); i++) {
                // Start by saying there should be no switching:
                shouldSwitch = false;
                // Get the two elements you want to compare, one from current row and one from the next:
                x = rows[i].getElementsByTagName("TD")[column];
                y = rows[i + 1].getElementsByTagName("TD")[column];
                // Check if the two rows should switch place, based on the direction, asc or desc:
                if (dir == "asc") {
                if (x.innerHTML.toLowerCase() > y.innerHTML.toLowerCase()) {
                    // If so, mark as a switch and break the loop:
                    shouldSwitch= true;
                    break;
                }
                } else if (dir == "desc") {
                if (x.innerHTML.toLowerCase() < y.innerHTML.toLowerCase()) {
                    // If so, mark as a switch and break the loop:
                    shouldSwitch = true;
                    break;
                }
                }
            }
            if (shouldSwitch) {
                // If a switch has been marked, make the switch and mark that a switch has been done:
                rows[i].parentNode.insertBefore(rows[i + 1], rows[i]);
                switching = true;
                // Each time a switch is done, increase this count by 1:
                switchcount ++;
            } else {
                // If no switching has been done AND the direction is "asc", set the direction to "desc" and run the while loop again.
                if (switchcount == 0 && dir == "asc") {
                dir = "desc";
                switching = true;
                }
            }
            }
        }

        // Code for generating Proksee projects.

        function loadScript(scriptUrl, callback) {
            const existingScript = document.querySelector(`script[src="${scriptUrl}"]`);
            if (existingScript) {
                existingScript.remove();
            }
            const script = document.createElement('script');
            script.src = scriptUrl;
            script.type = 'text/javascript';
            document.body.appendChild(script);
            script.onload = function() {
                console.log(`Script loaded: ${scriptUrl}`);
                if (typeof callback === "function") {
                    callback();
                }
            };
            script.onerror = function() {
                console.error(`Error loading script: ${scriptUrl}`);
                alert(`Failed to load data for script: ${scriptUrl}`);
            };
        }

        function generateProkseeLink(element, sampleId) {
            loadScript(`${sampleId}.js`, function() {
                if (typeof window.json === 'undefined') {
                    console.error('No data found for sample ID:', sampleId);
                    alert('Failed to generate Proksee project: No data available for this sample.');
                    return;
                }

                const data = { origin: 'proksee-batch', data: JSON.stringify(window.json) };
                const url = 'https://proksee.ca/api/v1/projects.json';
                fetch(url, {
                    method: 'POST',
                    headers: { 'Content-Type': 'application/json' },
                    body: JSON.stringify(data)
                })
                .then(response => response.json())
                .then(data => {
                    if (data?.status === 'success' && data?.url) {
                        const link = document.createElement('a');
                        link.textContent = 'Go to Proksee Project';
                        link.href = data.url;
                        link.className = 'generated-link';
                        link.target = '_blank';
                        element.parentNode.replaceChild(link, element);
                    } else {
                        console.error(`Failed to create Proksee project: ${data?.error}`);
                    }
                })
                .catch(error => {
                    console.error('Error:', error);
                });
            });
        }

        function adjustTableHeight() {
            const headerHeight = document.querySelector('header').offsetHeight;
            const availableHeight = window.innerHeight - headerHeight;
            const scrollableTable = document.querySelector('.scrollable-table');
            scrollableTable.style.height = (availableHeight * 0.9) + 'px';
        }

        // Event listener for window resize
        window.addEventListener('resize', adjustTableHeight);

        // Initial adjustment on page load
        window.onload = function() {
            adjustTableHeight();
        };

        // Toggle Labels (not included in the controls.js file)
        onClick('btn-toggle-legend', () => {
        cgv.legend.update({visible: !cgv.legend.visible});
        cgv.draw();
        });

    </script>
</body>
</html>"""
        )



from typing import List
from typing import Tuple


def remove_covered_features(features: List[Tuple[int, int, int]]) -> List[bool]:
    """
    Identifies features that are completely covered by another feature with an
    equal or higher score.

    Parameters:
    features (list of tuples): A list of tuples, each containing a start position,
                               an end position, and a score.

    Returns:
    list: A list of boolean values indicating if the corresponding feature is not covered.
    """

    # Initialize an array to keep track of whether each feature is covered
    is_not_covered = [True] * len(features)

    # Iterate through each feature
    for i, (start, end, score) in enumerate(features):
        # Check for any other feature that covers this feature
        for j, (other_start, other_end, other_score) in enumerate(features):
            if i != j and other_score >= score:
                # Check if the other feature completely covers the current feature
                if other_start <= start and other_end >= end:
                    is_not_covered[i] = False
                    break

    return is_not_covered



import glob
import os
import time

from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.chrome.service import Service
from webdriver_manager.chrome import ChromeDriverManager


def setup_browser(download_dir: str) -> webdriver.Chrome:
    """
    Sets up the Chrome browser for headless operation.
    """
    chrome_options = Options()
    chrome_options.add_argument("--headless")
    chrome_options.add_argument("--no-sandbox")
    chrome_options.add_argument("--disable-dev-shm-usage")
    chrome_options.add_argument("--window-size=1920,1080")

    os.makedirs(download_dir, exist_ok=True)
    chrome_options.add_experimental_option(
        "prefs",
        {
            "download.default_directory": download_dir,
            "download.prompt_for_download": False,
            "download.directory_upgrade": True,
            "safebrowsing.enabled": True,
        },
    )

    service = Service(ChromeDriverManager().install())
    return webdriver.Chrome(service=service, options=chrome_options)  # type: ignore


def download_data(browser: webdriver.Chrome, url: str) -> None:
    """
    Navigates to the given URL and performs the downloading actions.
    """
    browser.get(url)
    time.sleep(5)

    try:
        first_button = browser.find_element("id", "react-tabs-4")
        first_button.click()
        time.sleep(3)

        second_button = browser.find_element(
            "css selector", "img[title='Download SVG']"
        )
        second_button.click()
        time.sleep(5)
    finally:
        browser.quit()


def scrape_proksee_image(proksee_link_file: str, output_file: str) -> None:
    """
    Scrapes the Proksee image from the given Proksee link file and saves it to the given output file.
    """
    # Define a temporary directory
    download_dir = os.path.join(
        os.path.dirname(output_file),
        "temporary_" + proksee_link_file.rsplit("/", 1)[-1],
    )

    # Extract link from file.
    proksee_link = None
    with open(proksee_link_file) as file:
        proksee_link = file.read().strip()

    # Set up the browser and download the data.
    browser = setup_browser(download_dir)
    download_data(browser, proksee_link)

    # Check that the download was successful.
    if len(glob.glob(os.path.join(download_dir, "*.svg"))) == 0:
        raise Exception("No .svg file was downloaded.")

    # Move the downloaded .svg file from the temporary directory to the output path.
    svg_file = glob.glob(os.path.join(download_dir, "*.svg"))[0]
    os.rename(svg_file, output_file)

    # Remove the temporary directory.
    os.rmdir(download_dir)



import json
from typing import Any
from typing import Dict


def merge_cgview_json_with_template(
    basic_json_file: str, template_file: str, output_file: str
) -> None:
    """
    Merge a basic cgview map in JSON format with a Proksee configuration file in JSON format.
    """
    # Load the basic cgview map
    with open(basic_json_file) as basic_json:
        basic_cgview_map = json.load(basic_json)

    # Load the Proksee configuration file
    with open(template_file) as template_json:
        proksee_configuration = json.load(template_json)

    # Define keys for formatting and data components
    formatting_keys = [
        "version",
        "created",
        "updated",
        "settings",
        "backbone",
        "ruler",
        "annotation",
        "dividers",
        "highlighter",
        "captions",
        "legend",
    ]
    data_keys = ["name", "sequence", "features", "plots", "bookmarks", "tracks"]

    # Create a new JSON structure
    merged_data: Dict[str, Any] = {"cgview": {}}

    # Add formatting components
    for key in formatting_keys:
        if key in proksee_configuration["cgview"]:
            merged_data["cgview"][key] = proksee_configuration["cgview"][key]

    # Add data components
    for key in data_keys:
        if key in basic_cgview_map["cgview"]:
            merged_data["cgview"][key] = basic_cgview_map["cgview"][key]

    # Assert that there is only one element in the "captions" list.
    assert (
        len(merged_data["cgview"]["captions"]) == 1
    ), "The 'captions' list must contain exactly one element."
    # Change the value of the "name" key in the first element of the "captions"
    # list to the value of the "name" key in the main "cgview" dictionary.
    merged_data["cgview"]["captions"][0]["name"] = merged_data["cgview"]["name"]

    # Write the merged data to a new file
    with open(output_file, "w") as file:
        json.dump(merged_data, file, indent=4)



import datetime
import json
import random
import string
from typing import Any
from typing import Dict
from typing import List

from Bio import SeqIO
from Bio.SeqFeature import CompoundLocation
from Bio.SeqFeature import FeatureLocation
from Bio.SeqRecord import SeqRecord


# Global variables (which should later be defined in a config file).
genetic_code: str = "11"
orientation: str = "+"


def genbank_to_cgview_json(genome_name: str, genbank_file: str, json_file: str) -> None:
    """Convert a GenBank file to a CGView JSON file."""
    genbank_records: List[SeqRecord] = list(SeqIO.parse(genbank_file, "genbank"))  # type: ignore

    now: datetime.datetime = datetime.datetime.now()

    map_id: str = "".join(random.choices(string.ascii_lowercase + string.digits, k=40))

    json_data: Dict[str, Any] = {
        "cgview": {
            "version": "1.0",
            "created": now.strftime("%Y-%m-%d %H:%M:%S"),
            "id": map_id,
            "name": genome_name,
            "geneticCode": genetic_code,
            "settings": {},
            "backbone": {},
            "ruler": {},
            "dividers": {},
            "annotation": {},
            "sequence": {"contigs": []},
            "captions": [],
            "legend": {"items": []},
            "features": [],
            "tracks": [
                {
                    "name": "Features",
                    "separateFeaturesBy": "strand",
                    "position": "both",
                    "dataType": "feature",
                    "dataMethod": "source",
                    "dataKeys": "genbank-features",
                }
            ],
        }
    }

    all_contig_names: List[str] = []
    for record in genbank_records:
        # Determine contig name based on availability of attributes.
        genbank_contig_name = ""
        if record.id is not None and record.id:
            genbank_contig_name = record.id
        elif record.name is not None and record.name:
            genbank_contig_name = record.name

        assert (
            genbank_contig_name not in all_contig_names
        ), f"Duplicate contig name {genbank_contig_name} in {genbank_file}."
        all_contig_names.append(genbank_contig_name)

        contig_name = remove_problematic_characters_from_contig_name(
            genbank_contig_name
        )

        contig_data = seq_to_json_contig(contig_name, str(record.seq))

        json_data["cgview"]["sequence"]["contigs"].append(contig_data)

        features_to_skip = ["source", "gene", "exon"]

        for feature in record.features:
            if feature.type in features_to_skip:
                continue

            if feature.location is None:
                raise ValueError("Location is None")

            # If the feature is a CDS, determine the reading frame.
            feature_codon_start = 1
            if feature.type == "CDS":
                codon_start_qualifier = feature.qualifiers.get("codon_start")
                if codon_start_qualifier is not None and codon_start_qualifier:
                    feature_codon_start = int(codon_start_qualifier[0])
            assert feature_codon_start in [
                1,
                2,
                3,
            ], f"codon_start must be 1, 2, or 3, not {feature_codon_start}"

            # Determine the location type, and process accordingly
            locations = (
                feature.location.parts
                if isinstance(feature.location, CompoundLocation)
                else [feature.location]
                if isinstance(feature.location, FeatureLocation)
                else None
            )

            # Determine strand.
            strand = 0
            if {-1} == {loc.strand for loc in locations}:
                strand = -1
            elif {1} == {loc.strand for loc in locations}:
                strand = 1
            else:
                raise ValueError(
                    "Strand is not consistent among locations within feature {}".format(
                        feature
                    )
                )

            if feature.type == "CDS":
                if strand == 1:
                    # Assume that the locations are already sorted by start position from 5' to 3' on the forward strand.
                    pass
                elif strand == -1:
                    # Sort the locations by start position from 3' to 5' on the reverse strand by reversing the list.
                    locations = list(reversed(locations))

            # Initiate a counter for the number of base pairs from the start of
            # the first codon (for use in the case of CDS features).
            cumulative_bp_from_first_codon_start = 0

            # Iterate over the locations within this feature.
            loc_num = 0
            for loc in locations:
                loc_num += 1

                # Switch back to 1-based indexing (Biopython converts to 0-based indexing when parsing GenBank format).
                start = int(loc.start) + 1
                stop = int(loc.end)

                # Assign name based on availability of attributes.
                name = ""
                locus_tag = feature.qualifiers.get("locus_tag")
                gene = feature.qualifiers.get("gene")
                product = feature.qualifiers.get("product")

                if locus_tag is not None and locus_tag:
                    name = locus_tag[0]
                elif gene is not None and gene:
                    name = gene[0]
                elif product is not None and product:
                    name = product[0]

                codon_start = 1
                if feature.type == "CDS":
                    # Figure out what the reading frame is for this location
                    # within the CDS feature.
                    if loc_num == 1:
                        # The codon start for this location is the same as the
                        # codon start for the original CDS feature.
                        codon_start = feature_codon_start

                        # Start counting at the first bp of the first codon.
                        cumulative_bp_from_first_codon_start = (
                            abs(stop - start) + 1 - (feature_codon_start - 1)
                        )

                    else:
                        remainder_after_last_full_codon = (
                            cumulative_bp_from_first_codon_start % 3
                        )
                        if remainder_after_last_full_codon == 0:
                            codon_start = 1
                        elif remainder_after_last_full_codon == 1:
                            codon_start = 3
                        elif remainder_after_last_full_codon == 2:
                            codon_start = 2

                        # Add the length of the current location to the total.
                        cumulative_bp_from_first_codon_start += abs(stop - start) + 1

                feature_data = {
                    "type": feature.type,
                    "name": name,
                    "start": start,
                    "stop": stop,
                    "strand": strand,
                    "source": "genbank-features",
                    "contig": contig_name,
                    "legend": feature.type,
                }

                if feature.type == "CDS":
                    # Define the reading frame in the feature data using the JSON feature "codonStart".
                    feature_data["codonStart"] = codon_start

                json_data["cgview"]["features"].append(feature_data)

    with open(json_file, "w") as outfile:
        json.dump(json_data, outfile, indent=4)


def fasta_to_cgview_json(genome_name: str, fasta_file: str, json_file: str) -> None:
    """Convert a FASTA file to a CGView JSON file. The JSON file will be in the
    same format as generated by the `genbank_to_cgview_json` function. There
    will be no features in the JSON file, only sequences/contigs.
    """
    fasta_records: List[SeqRecord] = list(SeqIO.parse(fasta_file, "fasta"))  # type: ignore

    now: datetime.datetime = datetime.datetime.now()

    map_id: str = "".join(random.choices(string.ascii_lowercase + string.digits, k=40))

    json_data: Dict[str, Any] = {
        "cgview": {
            "version": "1.0",
            "created": now.strftime("%Y-%m-%d %H:%M:%S"),
            "id": map_id,
            "name": genome_name,
            "geneticCode": genetic_code,
            "settings": {},
            "backbone": {},
            "ruler": {},
            "dividers": {},
            "annotation": {},
            "sequence": {"contigs": []},
            "captions": [],
            "legend": {"items": []},
            "features": [],
            "tracks": [],
        }
    }

    all_contig_names: List[str] = []

    contigs_json = []
    for record in fasta_records:
        # Determine contig name based on availability of attributes.
        fasta_contig_name = ""
        if record.id is not None and record.id:
            fasta_contig_name = record.id
        elif record.description is not None and record.description:
            fasta_contig_name = record.description

        assert (
            fasta_contig_name not in all_contig_names
        ), f"Duplicate contig name {fasta_contig_name} in {fasta_file}."
        all_contig_names.append(fasta_contig_name)

        contig_name = remove_problematic_characters_from_contig_name(fasta_contig_name)

        contigs_json.append(seq_to_json_contig(contig_name, str(record.seq)))

    json_data["cgview"]["sequence"]["contigs"] = contigs_json

    with open(json_file, "w") as json_fh:
        json.dump(json_data, json_fh, indent=4)


def seq_to_json_contig(seq_id: str, seq: str) -> Dict[str, Any]:
    """Convert a sequence to a dictionary with the sequence ID, sequence length,
    and sequence. The dictionary will be in the format expected for a sequence
    in a CGView JSON file.
    """
    # Define a dictionary to hold the sequence information.
    seq_dict = {}

    # Add the sequence ID to the dictionary.
    seq_dict["name"] = seq_id

    # Add the sequence length to the dictionary.
    seq_dict["length"] = str(len(seq))

    # Add the sequence to the dictionary.
    seq_dict["seq"] = seq

    # Add the sequence orientation to the dictionary.
    seq_dict["orientation"] = orientation

    # Return the dictionary.
    return seq_dict


def remove_problematic_characters_from_contig_name(contig_name: str) -> str:
    """Remove problematic characters from a contig name. This is necessary
    some downstream software may not be able to handle contig names with
    certain characters.
    """
    return contig_name.replace("|", "_").replace(";", "_")



import json

import requests


def generate_proksee_link(json_data_file: str, output_file: str) -> None:
    """
    Generate a Proksee link from a JSON file.
    """
    url = "https://proksee.ca/api/v1/projects.json"

    json_data = None
    with open(json_data_file) as file:
        json_data = json.load(file)
    data = {"data": json.dumps(json_data), "origin": "proksee-batch"}

    try:
        response = requests.post(
            url, headers={"Content-Type": "application/json"}, json=data
        )
        response.raise_for_status()
        data = response.json()
        if data.get("status") == "success" and data.get("url"):
            with open(output_file, "w") as file:
                file.write(data["url"])
        else:
            print(f"Failed to create Proksee project: {data.get('error')}")
    except requests.RequestException as e:
        print(f"Error: {e}")



"""

The input directory must be structured as in the following example:

    input_directory/
        genome_name_1/
            genbank/
                genome1.gbk
            fasta/
                genome1.fna
            blast/
                abc.txt
                def.tsv
            bed/
                ghi.bed
                jkl.bed
            json/
                template1.json
            vcf/
                mno.vcf
                pqr.vcf
            gff/
                stu.gff
                vwx.gff3
        genome_name_2/
            genbank/
                genome2.gbff
            fasta/
                genome2.fa
            blast/
                yza.txt
                bcd.tsv
            bed/
                efg.bed
                hij.bed
            json/
                template2.json
            vcf/
                klm.vcf
                nop.vcf
            gff/
                qrs.gff
                tuv.gff3
        ...

The genbank directory must contain a single GenBank file with the extension
.gbk, .gbff, or .gb. This is the genome that will be visualized. If the genbank
directory is not present, then proksee-batch will use a file from the fasta
directory instead (otherwise the fasta directory is ignored).  The blast, bed,
vcf, and gff directories are optional. They contain files with additional
genomic features.  The json directory is also optional. It contains a custom
Proksee project JSON file that will be used as a template for the visualization.
"""
import json
import os
import sys
from typing import Dict
from typing import List

from Bio import SeqIO


def validate_input_directory_contents(input: str) -> None:
    """
    Validates if the provided input directory contains the required subdirectories and files.

    Args:
        input (str): The path to the input directory to be checked.

    Raises:
        SystemExit: If the input directory does not contain the required subdirectories.
    """
    # Check that the input directory exists.
    if not os.path.isdir(input):
        handle_error_exit(f"The input directory {input} does not exist.")

    # Iterate over subdirectories in the input directory (each subdirectory
    # should contain the data for one genome).
    for genome_dir in os.listdir(input):
        # Check that the subdirectory contains a genbank subdirectory, a fasta subdirectory, or both.
        required_subdirectories = ["genbank", "fasta"]
        if not any(
            os.path.isdir(os.path.join(input, genome_dir, subdirectory))
            for subdirectory in required_subdirectories
        ):
            handle_error_exit(
                f"The input directory {input} does not contain any of the required subdirectories: {required_subdirectories}"
            )

        # Check that all subdirectories have valid names.
        valid_subdirectories = [
            "genbank",
            "fasta",
            "blast",
            "bed",
            "json",
            "vcf",
            "gff",
        ]
        for subdirectory in os.listdir(os.path.join(input, genome_dir)):
            if subdirectory not in valid_subdirectories:
                handle_error_exit(
                    f"The input directory contains an invalid subdirectory: {subdirectory}"
                )

        # Check the contents of the subdirectories, if they exist.
        seq_file_path = ""
        seq_file_format = ""
        for subdirectory in valid_subdirectories:
            subdirectory_path = os.path.join(input, genome_dir, subdirectory)
            if os.path.isdir(subdirectory_path):
                if subdirectory == "genbank":
                    genbank_files = []
                    for file in os.listdir(subdirectory_path):
                        if (
                            file.endswith(".gbk")
                            or file.endswith(".gbff")
                            or file.endswith(".gb")
                        ):
                            genbank_files.append(file)
                            # Parse the GenBank file with BioPython to check if the file format is valid.
                            try:
                                SeqIO.parse(
                                    os.path.join(subdirectory_path, file), "genbank"
                                ).__next__()  # type: ignore
                            except Exception as e:
                                handle_error_exit(
                                    f"Error parsing the GenBank file {file}: {e}. Please check that the file is in GenBank format."
                                )
                    if len(genbank_files) == 0:
                        handle_error_exit(
                            f"The input directory {subdirectory_path} does not contain any GenBank files. Valid file extensions are .gbk, .gbff, and .gb."
                        )
                    elif len(genbank_files) > 1:
                        handle_error_exit(
                            f"The input directory {subdirectory_path} contains more than one GenBank file. Please ensure that there is exactly one GenBank file in the directory."
                        )
                    seq_file_path = os.path.join(subdirectory_path, genbank_files[0])
                    seq_file_format = "genbank"
                elif subdirectory == "fasta":
                    fasta_files = []
                    for file in os.listdir(subdirectory_path):
                        if (
                            file.endswith(".fna")
                            or file.endswith(".fa")
                            or file.endswith(".fasta")
                        ):
                            fasta_files.append(file)
                            # Parse the FASTA file with BioPython to check if the file format is valid.
                            try:
                                SeqIO.parse(
                                    os.path.join(subdirectory_path, file), "fasta"
                                ).__next__()  # type: ignore
                            except Exception as e:
                                handle_error_exit(
                                    f"Error parsing the FASTA file {file}: {e}. Please check that the file is in FASTA format."
                                )
                            # Check that the sequences are nucleotide sequences, and not protein sequences.
                            with open(os.path.join(subdirectory_path, file)) as f:
                                for line in f:
                                    if not line.startswith(">") and line.strip() != "":
                                        if not set(line.strip().upper()).issubset(
                                            set("ATGCNatgcn")
                                        ):
                                            handle_error_exit(
                                                f"The FASTA file {file} contains one-letter codes that do not represent DNA nucleotide bases. Please check that the file contains nucleotide sequences only."
                                            )
                    if len(fasta_files) == 0:
                        handle_error_exit(
                            f"The input directory {subdirectory_path} does not contain any FASTA files. Valid file extensions are .fna and .fa."
                        )
                    elif len(fasta_files) > 1:
                        handle_error_exit(
                            f"The input directory {subdirectory_path} contains more than one FASTA file. Please ensure that there is exactly one FASTA file in the directory."
                        )
                    if seq_file_path == "":
                        seq_file_path = os.path.join(subdirectory_path, fasta_files[0])
                        seq_file_format = "fasta"
                elif subdirectory == "blast":
                    blast_result_files = []
                    for file in os.listdir(subdirectory_path):
                        if file.endswith(".txt") or file.endswith(".tsv"):
                            blast_result_files.append(file)
                            # Validate the BLAST result file. It should be in tabular format (outfmt 6).
                            with open(os.path.join(subdirectory_path, file)) as f:
                                for line in f:
                                    if not line.startswith("#"):
                                        line_split = line.split("\t")
                                        if len(line_split) != 12:
                                            handle_error_exit(
                                                f"The BLAST result file {file} does not have the correct number of columns. Please check that the file is in tabular format (outfmt 6)."
                                            )
                                        try:
                                            float(line_split[2])
                                            float(line_split[3])
                                            float(line_split[10])
                                            float(line_split[11])
                                        except ValueError:
                                            handle_error_exit(
                                                f"The BLAST result file {file} does not have the correct format. Please check that the file is in tabular format (outfmt 6)."
                                            )
                    if len(blast_result_files) == 0:
                        handle_error_exit(
                            f"The input directory {subdirectory_path} does not contain any BLAST result files. Valid file extensions are .txt and .tsv."
                        )
                elif subdirectory == "bed":
                    bed_files = []
                    for file in os.listdir(subdirectory_path):
                        if file.endswith(".bed"):
                            bed_files.append(file)
                            # Validate the BED file. It should be in BED format.
                            with open(os.path.join(subdirectory_path, file)) as f:
                                for line in f:
                                    line_split = line.split("\t")
                                    if len(line_split) < 3:
                                        handle_error_exit(
                                            f"The BED file {file} does not have the correct number of columns. Please check that the file is in BED format."
                                        )
                                    try:
                                        int(line_split[1])
                                        int(line_split[2])
                                    except ValueError:
                                        handle_error_exit(
                                            f"The BED file {file} does not have the correct format. Please check that the file is in BED format."
                                        )
                    if len(bed_files) == 0:
                        handle_error_exit(
                            f"The input directory {subdirectory_path} does not contain any BED files. Valid file extension is .bed."
                        )
                elif subdirectory == "json":
                    json_files = []
                    for file in os.listdir(subdirectory_path):
                        if file.endswith(".json"):
                            json_files.append(file)
                            # Validate the JSON file. It should be in JSON format and contain the required keys.
                            with open(os.path.join(subdirectory_path, file)) as f:
                                try:
                                    json_dict = json.load(f)
                                except Exception as e:
                                    handle_error_exit(
                                        f"Error parsing the JSON file {file}: {e}. Please check that the file is in JSON format."
                                    )
                                required_keys = [
                                    "cgview",
                                ]
                                for key in required_keys:
                                    if key not in json_dict:
                                        handle_error_exit(
                                            f"The JSON file {file} does not contain the required key: {key}"
                                        )
                    if len(json_files) == 0:
                        handle_error_exit(
                            f"The input directory {subdirectory_path} does not contain any JSON files. Valid file extension is .json."
                        )
                elif subdirectory == "vcf":
                    vcf_files = []
                    for file in os.listdir(subdirectory_path):
                        if file.endswith(".vcf"):
                            vcf_files.append(file)
                            # Validate the VCF file. It should be in VCF format.
                            with open(os.path.join(subdirectory_path, file)) as f:
                                for line in f:
                                    if not line.startswith("#"):
                                        line_split = line.split("\t")
                                        if len(line_split) != 8:
                                            handle_error_exit(
                                                f"The VCF file {file} does not have the correct number of columns. Please check that the file is in VCF format."
                                            )
                                        try:
                                            int(line_split[1])
                                            try:
                                                float(line_split[5])
                                            except:
                                                assert line_split[5] == "."
                                        except ValueError:
                                            handle_error_exit(
                                                f"The VCF file {file} does not have the correct format. Please check that the file is in VCF format."
                                            )
                            # Check that all sequence IDs in the VCF file are contigs in the GenBank file.
                            if not check_vcf_seq_ids(
                                os.path.join(subdirectory_path, file),
                                seq_file_path,
                                seq_file_format,
                            ):
                                handle_error_exit(
                                    f"The VCF file {file} contains sequence IDs that are not contigs in the GenBank file {genbank_files[0]}."
                                )
                            # Check that the genotypes in the genome in the GenBank file match the REF genotypes in the VCF file.
                            if not check_vcf_ref_vs_alt_genotypes(
                                os.path.join(subdirectory_path, file),
                                seq_file_path,
                                seq_file_format,
                            ):
                                handle_error_exit(
                                    f"The genotypes in the genome in the GenBank file {genbank_files[0]} do not match the REF genotypes in the VCF file {file}."
                                )
                    if len(vcf_files) == 0:
                        handle_error_exit(
                            f"The input directory {subdirectory_path} does not contain any VCF files. Valid file extension is .vcf."
                        )
                elif subdirectory == "gff":
                    gff_files = []
                    for file in os.listdir(subdirectory_path):
                        if file.endswith(".gff") or file.endswith(".gff3"):
                            gff_files.append(file)
                            # Validate the GFF file. It should be in GFF format.
                            with open(os.path.join(subdirectory_path, file)) as f:
                                for line in f:
                                    if not line.startswith("#"):
                                        line_split = line.split("\t")
                                        if len(line_split) != 9:
                                            handle_error_exit(
                                                f"The GFF file {file} does not have the correct number of columns. Please check that the file is in GFF format."
                                            )
                                        try:
                                            int(line_split[3])
                                            int(line_split[4])
                                            try:
                                                float(line_split[5])
                                            except:
                                                assert line_split[5] == "."
                                        except ValueError:
                                            handle_error_exit(
                                                f"The GFF file {file} does not have the correct format. Please check that the file is in GFF format."
                                            )
                    if len(gff_files) == 0:
                        handle_error_exit(
                            f"The input directory {subdirectory_path} does not contain any GFF files. Valid file extensions are .gff and .gff3."
                        )


def get_data_files(input_subdir: str, data_type: str) -> List[str]:
    """
    Returns the paths to the data files of the specified type in the provided subdirectory.

    Args:
        input_subdir (str): The path to the subdirectory containing the data files.
        data_type (str): The type of the data files to be returned. Valid values are "genbank", "blast", "bed", "json", "vcf", and "gff".

    Returns:
        list: The paths to the data files.
    """
    data_dir = os.path.join(input_subdir, data_type)
    if os.path.isdir(data_dir):
        if data_type == "genbank":
            genbank_paths = [
                os.path.join(data_dir, file)
                for file in os.listdir(data_dir)
                if file.endswith(".gbk")
                or file.endswith(".gbff")
                or file.endswith(".gb")
            ]
            assert (
                len(genbank_paths) == 1
            ), f"The input directory {data_dir} does not contain exactly one GenBank file. Valid file extensions are .gbk, .gbff, and .gb."
            return genbank_paths
        elif data_type == "fasta":
            fasta_paths = [
                os.path.join(data_dir, file)
                for file in os.listdir(data_dir)
                if file.endswith(".fna")
                or file.endswith(".fa")
                or file.endswith(".fasta")
            ]
            assert (
                len(fasta_paths) == 1
            ), f"The input directory {data_dir} does not contain exactly one FASTA file. Valid file extensions are .fna, .fa, and .fasta."
            return fasta_paths
        elif data_type == "blast":
            return [
                os.path.join(data_dir, file)
                for file in os.listdir(data_dir)
                if file.endswith(".txt") or file.endswith(".tsv")
            ]
        elif data_type == "bed":
            return [
                os.path.join(data_dir, file)
                for file in os.listdir(data_dir)
                if file.endswith(".bed")
            ]
        elif data_type == "json":
            return [
                os.path.join(data_dir, file)
                for file in os.listdir(data_dir)
                if file.endswith(".json")
            ]
        elif data_type == "vcf":
            return [
                os.path.join(data_dir, file)
                for file in os.listdir(data_dir)
                if file.endswith(".vcf")
            ]
        elif data_type == "gff":
            return [
                os.path.join(data_dir, file)
                for file in os.listdir(data_dir)
                if file.endswith(".gff") or file.endswith(".gff3")
            ]
        else:
            handle_error_exit(f"Invalid data type: {data_type}")
    return []


def handle_error_exit(error_message: str, exit_code: int = 1) -> None:
    """
    Handles errors by printing a message to sys.stderr and exiting the program.

    Args:
        error_message (str): The error message to be printed.
        exit_code (int): The exit code to be used for sys.exit. Defaults to 1.

    Exits:
        SystemExit: Exits the program with the provided exit code.
    """
    print(error_message, file=sys.stderr)
    sys.exit(exit_code)


def check_vcf_seq_ids(
    vcf_file_path: str, seq_file_path: str, seq_file_format: str
) -> bool:
    """
    Checks if all the sequence IDs in the first column of the VCF file are contigs in a GenBank or FASTA file.

    Args:
        vcf_file_path (str): The path to the VCF file.
        seq_file_path (str): The path to the GenBank or FASTA file.
        seq_file_format (str): The format of the GenBank or FASTA file. Valid values are "genbank" and "fasta".

    Returns:
        bool: True if all sequence IDs in the VCF file are contigs in the sequence file, False otherwise.
    """
    assert seq_file_format in ["genbank", "fasta"]
    vcf_seq_ids = []
    with open(vcf_file_path) as vcf_file:
        for line in vcf_file:
            if not line.startswith("#") and line.strip() != "":
                vcf_seq_ids.append(line.split("\t")[0])
    seq_ids = []
    for record in SeqIO.parse(seq_file_path, seq_file_format):  # type: ignore
        seq_ids.append(record.id)
    return all(seq_id in seq_ids for seq_id in vcf_seq_ids)


def check_vcf_ref_vs_alt_genotypes(
    vcf_file_path: str, genome_file_path: str, genome_file_type: str
) -> bool:
    """
    Checks if the genotypes in the genome in the GenBank file match the REF genotypes in the VCF file.
    Args:
        vcf_file_path (str): The path to the VCF file.
        genbank_file_path (str): The path to the GenBank file.
    Returns:
        bool: True if the genotypes in the genome in the GenBank file match the REF genotypes in the VCF file, False otherwise.
    """
    # Get the genome sequences from the GenBank file
    genome_sequences = {}
    if genome_file_type == "genbank":
        genome_sequences = {
            record.id: str(record.seq)
            for record in SeqIO.parse(genome_file_path, "genbank")  # type: ignore
        }
    elif genome_file_type == "fasta":
        genome_sequences = {
            record.id: str(record.seq)
            for record in SeqIO.parse(genome_file_path, "fasta")  # type: ignore
        }
    else:
        handle_error_exit(f"Invalid genome file type: {genome_file_type}")

    # Parse the VCF file content
    vcf_data = []
    with open(vcf_file_path) as vcf_file:
        for line in vcf_file:
            if not line.startswith("#") and line.strip():
                parts = line.split("\t")
                vcf_data.append((parts[0], int(parts[1]), parts[3]))

    # Check each genotype in the VCF file
    for chrom, pos, ref_genotype in vcf_data:
        actual_genotype = genome_sequences[chrom][pos - 1]
        if ref_genotype != actual_genotype:
            print(
                f"Genotype mismatch at {chrom}:{pos}: {actual_genotype} should be {ref_genotype}"
            )
            return False

    return True



"""Main module for the Proksee Batch tool.

This file defines the proksee-batch command-line interface and implements the
high-level logic of the tool.

"""
import json
import os
import shutil
import sys
from importlib import resources
from importlib.metadata import version
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

import click
import toml

from .download_example_input import download_example_input

# from .generate_proksee_link import generate_proksee_link
from .generate_report_html import generate_report_html
from .get_stats_from_seq_file import get_stats_from_seq_file
from .merge_cgview_json_with_template import merge_cgview_json_with_template
from .parse_additional_features import add_bed_features_and_tracks
from .parse_additional_features import add_blast_features_and_tracks
from .parse_additional_features import add_gff_features_and_tracks
from .parse_additional_features import add_vcf_features_and_tracks
from .seq_file_to_cgview_json import fasta_to_cgview_json
from .seq_file_to_cgview_json import genbank_to_cgview_json

# from .scrape_proksee_image import scrape_proksee_image
from .validate_input_data import get_data_files
from .validate_input_data import validate_input_directory_contents


@click.command()
@click.version_option()
@click.option(
    "--input",
    type=click.Path(exists=True, file_okay=False, dir_okay=True),
    help="Path to the directory containing input genomes in GenBank format.",
)
@click.option(
    "--output",
    type=click.Path(file_okay=False, dir_okay=True),
    help="Path to the output directory.",
)
@click.option(
    "--download-example-data",
    type=click.Path(file_okay=False, dir_okay=True),
    help="Download example input files to the specified directory, and exit.",
)
def main(
    input: Optional[str],
    output: Optional[str],
    download_example_data: Optional[str],
) -> None:
    """
    Proksee Batch: A tool for visualizing batches of genomes via https://www.proksee.ca.
    """

    # Check if the download example GenBank files option is used
    if download_example_data:
        # Download logic
        download_example_input(download_example_data)
        print(f"Example data files downloaded to {download_example_data}")
        sys.exit(0)

    # Check if the input directory is provided.
    input_dir_path = ""
    if not input:
        handle_error_exit("Missing option '--input'.")
    else:
        assert input is not None
        if not os.path.exists(input):
            handle_error_exit(f"The input directory does not exist: {input}")
        input_dir_path = os.path.abspath(input)

    # Validate the contents of the input directory.
    validate_input_directory_contents(input_dir_path)

    # Make the output directory if it doesn't exist.
    output_path = ""
    if not output:
        handle_error_exit("Missing option '--output'.")
    else:
        output_path = os.path.abspath(output)
        if os.path.exists(output_path):
            shutil.rmtree(output_path)
        os.mkdir(output_path)
        os.mkdir(os.path.join(output_path, "data"))
        # os.mkdir(os.path.join(output_path, "images"))

    # Initiate a dictionary to store file paths for each genome.
    genome_info: Dict[str, Dict[str, Any]] = {}

    # Iterate over subdirectories in the input directory.
    genome_num = 0
    for genome_dir in os.listdir(input):
        # Skip any files in the input directory.
        if os.path.isfile(os.path.join(input_dir_path, genome_dir)):
            continue

        genome_num += 1
        genome_code_name = "genome_" + str(genome_num)

        # Define path to a temporary output directory within the output directory.
        temp_output = os.path.join(output_path, genome_code_name + "_temp")
        os.mkdir(temp_output)

        # Get the data file paths
        genbank_paths = get_data_files(
            os.path.join(input_dir_path, genome_dir), "genbank"
        )
        fasta_paths = get_data_files(os.path.join(input_dir_path, genome_dir), "fasta")
        json_paths = get_data_files(os.path.join(input_dir_path, genome_dir), "json")
        blast_paths = get_data_files(os.path.join(input_dir_path, genome_dir), "blast")
        bed_paths = get_data_files(os.path.join(input_dir_path, genome_dir), "bed")
        vcf_paths = get_data_files(os.path.join(input_dir_path, genome_dir), "vcf")
        gff_paths = get_data_files(os.path.join(input_dir_path, genome_dir), "gff")

        # Define path to the basic cgview map in JSON format.
        basic_json_file = os.path.join(temp_output, genome_code_name + ".json")

        if genbank_paths:
            genbank_path = genbank_paths[0]

            # Get basic stats from the GenBank file.
            (
                genbank_description,
                genbank_total_size,
                genbank_number_of_contigs,
                genbank_gc_content,
            ) = get_stats_from_seq_file(genbank_path, "genbank")

            genome_info[genome_code_name] = {
                "Name": genome_dir,
                "Description": genbank_description,
                "Total size": genbank_total_size,
                "Number of contigs": genbank_number_of_contigs,
                "GC content": genbank_gc_content,
            }

            # Convert the GenBank file to a basic cgview map in JSON format.
            genbank_to_cgview_json(genome_dir, genbank_path, basic_json_file)

        elif fasta_paths:
            fasta_path = fasta_paths[0]

            # Get basic stats from the FASTA file.
            (
                fasta_description,
                fasta_total_size,
                fasta_number_of_contigs,
                fasta_gc_content,
            ) = get_stats_from_seq_file(fasta_path, "fasta")

            genome_info[genome_code_name] = {
                "Name": genome_dir,
                "Description": fasta_description,
                "Total size": fasta_total_size,
                "Number of contigs": fasta_number_of_contigs,
                "GC content": fasta_gc_content,
            }

            # Convert the FASTA file to a basic cgview map in JSON format.
            fasta_to_cgview_json(genome_dir, fasta_path, basic_json_file)

        else:
            handle_error_exit(
                f"Could not find a GenBank or FASTA file in the {genome_dir} directory."
            )

        # Parse the BLAST result files (if any) to create additional features and tracks.
        basic_json_file_with_blast_features = os.path.join(
            temp_output, genome_code_name + ".with_blast_features.json"
        )
        if blast_paths:
            add_blast_features_and_tracks(
                blast_paths, basic_json_file, basic_json_file_with_blast_features
            )
        else:
            shutil.copy(basic_json_file, basic_json_file_with_blast_features)

        # Parse the BED files (if any) to create additional features and tracks.
        basic_json_file_with_bed_features = os.path.join(
            temp_output, genome_code_name + ".with_bed_features.json"
        )
        if bed_paths:
            add_bed_features_and_tracks(
                bed_paths,
                basic_json_file_with_blast_features,
                basic_json_file_with_bed_features,
            )
        else:
            shutil.copy(
                basic_json_file_with_blast_features,
                basic_json_file_with_bed_features,
            )

        # Parse the VCF files (if any) to create additional features and tracks.
        basic_json_file_with_vcf_features = os.path.join(
            temp_output, genome_code_name + ".with_vcf_features.json"
        )
        if vcf_paths:
            add_vcf_features_and_tracks(
                vcf_paths,
                basic_json_file_with_bed_features,
                basic_json_file_with_vcf_features,
            )
        else:
            shutil.copy(
                basic_json_file_with_bed_features,
                basic_json_file_with_vcf_features,
            )

        # Parse the GFF files (if any) to create additional features and tracks.
        basic_json_file_with_gff_features = os.path.join(
            temp_output, genome_code_name + ".with_gff_features.json"
        )
        if gff_paths:
            add_gff_features_and_tracks(
                gff_paths,
                basic_json_file_with_vcf_features,
                basic_json_file_with_gff_features,
            )
        else:
            shutil.copy(
                basic_json_file_with_vcf_features,
                basic_json_file_with_gff_features,
            )

        # Determine path to the template Proksee configuration file.
        template_path = ""
        if json_paths:
            try:
                with open(json_paths[0]) as template_file:
                    # Load the template file as a JSON object.
                    configuration = json.load(template_file)
                    template_path = json_paths[0]
            except json.JSONDecodeError as e:
                print(f"Error reading the template file: {e}", file=sys.stderr)
                sys.exit(1)
        else:
            with resources.path(
                "proksee_batch.data", "default_proksee_template.json"
            ) as template_json_path:
                template_path = str(template_json_path)
        assert os.path.exists(template_path)

        # Merge the basic cgview map with the template Proksee configuration file.
        merged_json_file = os.path.join(temp_output, genome_code_name + ".merged.json")
        merge_cgview_json_with_template(
            basic_json_file_with_gff_features, template_path, merged_json_file
        )

        # Add GC content and GC skew tracks to the merged JSON.
        merged_json_file_with_gc_tracks = os.path.join(
            temp_output, genome_code_name + ".merged_with_gc_tracks.json"
        )
        with open(merged_json_file) as merged_json_fh, open(
            merged_json_file_with_gc_tracks, "w"
        ) as merged_json_file_with_gc_tracks_fh:
            merged_json = json.load(merged_json_fh)
            merged_json["cgview"]["tracks"].extend(
                [
                    {
                        "name": "CG Content",
                        "thicknessRatio": 2,
                        "position": "inside",
                        "dataType": "plot",
                        "dataMethod": "sequence",
                        "dataKeys": "gc-content",
                    },
                    {
                        "name": "CG Skew",
                        "thicknessRatio": 2,
                        "position": "inside",
                        "dataType": "plot",
                        "dataMethod": "sequence",
                        "dataKeys": "gc-skew",
                    },
                ]
            )
            json.dump(merged_json, merged_json_file_with_gc_tracks_fh)

        # Convert the merged JSON file to .js file by wrapping it in a variable assignment.
        js_file = os.path.join(output_path, "data", genome_code_name + ".js")
        with open(js_file, "w") as file:
            # Get the JSON data from the merged JSON file as a string.
            json_data = None
            with open(merged_json_file_with_gc_tracks) as json_file:
                json_data = json_file.read()
            # Write the variable assignment.
            file.write(f"json = {json_data};")

        ## Generate a Proksee link using the merged JSON file.
        # proksee_project_link_file = os.path.join(
        #    temp_output, genome_code_name + ".proksee_link.txt"
        # )
        # generate_proksee_link(merged_json_file, proksee_project_link_file)

        ## Scrape proksee image.
        # proksee_image_file = os.path.join(
        #    os.path.join(output_path, "images", genome_code_name) + ".svg"
        # )
        # scrape_proksee_image(proksee_project_link_file, proksee_image_file)

        # Delete the temporary output directory.
        shutil.rmtree(temp_output)

    # Copy the directory with CGView.js code from the package data to the output directory.
    with resources.path("proksee_batch.data", "cgview-js_code") as cgview_js_path:
        shutil.copytree(cgview_js_path, os.path.join(output_path, "cgview-js_code"))

    # Generate the HTML report file.
    generate_report_html(output_path, genome_info)


def handle_error_exit(error_message: str, exit_code: int = 1) -> None:
    """
    Handles errors by printing a message to sys.stderr and exiting the program.

    Args:
        error_message (str): The error message to be printed.
        exit_code (int): The exit code to be used for sys.exit. Defaults to 1.

    Exits:
        SystemExit: Exits the program with the provided exit code.
    """
    print(error_message, file=sys.stderr)
    sys.exit(exit_code)


if __name__ == "__main__":
    main()



import json
import os
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple

import gffutils  # type: ignore

from .remove_covered_features import remove_covered_features


def parse_blast_files(
    blast_files: List[str],
) -> Tuple[List[Dict[str, Any]], List[Dict[str, Any]]]:
    """
    Parses BLAST result files.

    Parameters:
    blast_files (list): A list of paths to BLAST result files.

    Returns:
    tuple: A tuple containing a list of parsed BLAST features and a list of parsed BLAST tracks.
    """
    blast_features = []
    blast_tracks = []
    for i, blast_file in enumerate(blast_files):
        num = i + 1

        # Define the BLAST track.
        blast_track = {
            "name": os.path.basename(blast_file),
            "separateFeaturesBy": "none",
            "position": "both",
            "thicknessRatio": 1,
            "dataType": "feature",
            "dataMethod": "source",
            "dataKeys": f"blast_{num}",
            "drawOrder": "score",
        }
        # Add the BLAST track to the list of BLAST tracks.
        blast_tracks.append(blast_track)

        # Parse the BLAST result file.
        with open(blast_file) as f:
            blast_results = f.readlines()
        # Parse the BLAST result file.
        for blast_result_line in blast_results:
            # Parse the BLAST result line.
            blast_result = blast_result_line.split("\t")

            # Skip the header line(s), if present.
            if blast_result[0] == "qseqid" or blast_result[0].startswith("#"):
                continue

            # Define the BLAST feature.
            blast_feature = {
                "name": blast_result[
                    0
                ],  # Here we assume that the query sequence is a sequence aligned to a contig in the genome/map.
                "type": "blast",
                "start": int(
                    blast_result[8]
                ),  # Here we assume that the subject sequence is a contig in the genome/map.
                "stop": int(
                    blast_result[9]
                ),  # Here we assume that the subject sequence is a contig in the genome/map.
                "strand": 1 if blast_result[8] < blast_result[9] else -1,
                "source": f"blast_{num}",
                "contig": blast_result[
                    1
                ],  # Here we assume that the subject sequence is a contig in the genome/map.
                "legend": os.path.basename(blast_file),
                "tags": [],
                "meta": {
                    "identity": float(blast_result[2]),
                    "mismatches": int(blast_result[4]),
                    "evalue": float(blast_result[10]),
                    "bit_score": int(blast_result[11]),
                },
            }
            # Add the BLAST feature to the list of BLAST features.
            blast_features.append(blast_feature)

    # Remove BLAST features that are completely covered by another BLAST feature
    # with an equal or higher score.
    blast_features = [
        blast_feature
        for blast_feature, is_not_covered in zip(
            blast_features,
            remove_covered_features(
                get_feature_locations_and_scores_from_blast_features(blast_features)
            ),
        )
        if is_not_covered
    ]

    assert (
        len(blast_features) > 0 and len(blast_tracks) > 0
    ), "No BLAST features or tracks were obtained from input file {}.".format(
        blast_file
    )
    return (blast_features, blast_tracks)


def get_feature_locations_and_scores_from_blast_features(
    blast_features: List[Dict[str, Any]]
) -> List[Tuple[int, int, int]]:
    """
    Gets feature locations and scores from BLAST features.

    Parameters:
    blast_features (list): A list of parsed BLAST features.

    Returns:
    list: A list of tuples containing feature locations and scores.
    """
    feature_locations_and_scores = []
    for blast_feature in blast_features:
        feature_locations_and_scores.append(
            (
                blast_feature["start"],
                blast_feature["stop"],
                blast_feature["meta"]["bit_score"],
            )
        )
    return feature_locations_and_scores


def add_blast_features_and_tracks(
    blast_files: List[str], json_file: str, output_file: str
) -> None:
    """
    Parses BLAST result files, adds the parsed BLAST features and tracks to the cgview map JSON data structure, and writes the cgview map JSON data structure to a new file.

    Parameters:
    blast_files (list): A list of paths to BLAST result files.
    json_file (str): The path to a cgview map JSON file.
    output_file (str): The path to the output file.

    Returns:
    None
    """
    # Parse the BLAST result file.
    blast_features, blast_tracks = parse_blast_files(blast_files)
    # Read the cgview map JSON file.
    with open(json_file) as f:
        json_data = json.load(f)

    # Get list of contigs from JSON data.
    contigs = [contig["name"] for contig in json_data["cgview"]["sequence"]["contigs"]]

    # Check that all the blast features are assigned to contigs that are in the JSON data.
    for blast_feature in blast_features:
        assert (
            blast_feature["contig"] in contigs
        ), "The subject sequence {} listed in the BLAST result file {} is not among the contigs in the genome being mapped.".format(
            blast_feature["contig"], blast_feature["legend"]
        )

    # Add the parsed BLAST features and tracks to the cgview map JSON data structure.
    json_data["cgview"]["features"] += blast_features
    json_data["cgview"]["tracks"] += blast_tracks
    # Write the cgview map JSON data structure to a new file.
    with open(output_file, "w") as f:
        json.dump(json_data, f, indent=4)


def parse_bed_files(
    bed_files: List[str],
) -> Tuple[List[Dict[str, Any]], List[Dict[str, Any]]]:
    """
    Parses BED files.

    Parameters:
    bed_files (list): A list of paths to BED files.

    Returns:
    tuple: A tuple containing a list of parsed BED features and a list of parsed BED tracks.
    """
    bed_features = []
    bed_tracks = []
    for i, bed_file in enumerate(bed_files):
        num = i + 1

        # Define the BED track.
        bed_track = {
            "name": os.path.basename(bed_file),
            "separateFeaturesBy": "none",
            "position": "both",
            "thicknessRatio": 1,
            "dataType": "feature",
            "dataMethod": "source",
            "dataKeys": f"bed_{num}",
            "drawOrder": "score",
        }
        # Add the BED track to the list of BED tracks.
        bed_tracks.append(bed_track)

        # Parse the BED file.
        with open(bed_file) as f:
            bed_results = f.readlines()
        # Parse the BED file.
        for bed_result_line in bed_results:
            # Parse the BED result line.
            bed_result = bed_result_line.strip().split()

            # Skip the header line(s), if present.
            if (
                bed_result[0].startswith("#")
                or bed_result[0].startswith("track ")
                or bed_result[0].startswith("browser ")
            ):
                continue

            assert (
                len(bed_result) >= 3
            ), f"BED file {bed_file} is not in the correct format."

            name = ""
            if len(bed_result) >= 4:
                name = bed_result[3]
            score = 0
            if len(bed_result) >= 5:
                score = int(bed_result[4])
            strand = 1
            if len(bed_result) >= 6:
                if bed_result[5] == "+":
                    strand = 1
                elif bed_result[5] == "-":
                    strand = -1
                else:
                    raise ValueError(
                        "The strand column of BED file {} is not in the correct format.".format(
                            bed_file
                        )
                    )

            # Define the BED feature.
            bed_feature = {
                "name": name,
                "type": "bed",
                "start": int(bed_result[1]) + 1,
                "stop": int(bed_result[2]),
                "strand": strand,
                "source": f"bed_{num}",
                "contig": bed_result[0],
                "legend": os.path.basename(bed_file),
                "tags": [],
                "meta": {
                    "score": score,
                },
            }
            # Add the BED feature to the list of BED features.
            bed_features.append(bed_feature)

    # Remove BED features that are completely covered by another BED feature
    # with an equal or higher score.
    bed_features = [
        bed_feature
        for bed_feature, is_not_covered in zip(
            bed_features,
            remove_covered_features(
                get_feature_locations_and_scores_from_bed_features(bed_features)
            ),
        )
        if is_not_covered
    ]

    assert (
        len(bed_features) > 0 and len(bed_tracks) > 0
    ), f"No BED features or tracks were obtained from input file {bed_file}."
    return (bed_features, bed_tracks)


def get_feature_locations_and_scores_from_bed_features(
    bed_features: List[Dict[str, Any]]
) -> List[Tuple[int, int, int]]:
    """
    Gets feature locations and scores from BED features.

    Parameters:
    bed_features (list): A list of parsed BED features.

    Returns:
    list: A list of tuples containing feature locations and scores.
    """
    feature_locations_and_scores = []
    for bed_feature in bed_features:
        feature_locations_and_scores.append(
            (
                bed_feature["start"],
                bed_feature["stop"],
                bed_feature["meta"]["score"],
            )
        )
    return feature_locations_and_scores


def add_bed_features_and_tracks(
    bed_files: List[str], json_file: str, output_file: str
) -> None:
    """
    Parses BED files, adds the parsed BED features and tracks to the cgview map JSON data structure, and writes the cgview map JSON data structure to a new file.

    Parameters:
    bed_files (list): A list of paths to BED files.
    json_file (str): The path to a cgview map JSON file.
    output_file (str): The path to the output file.

    Returns:
    None
    """
    # Parse the BED file.
    bed_features, bed_tracks = parse_bed_files(bed_files)
    # Read the cgview map JSON file.
    with open(json_file) as f:
        json_data = json.load(f)

    # Get list of contigs from JSON data.
    contigs = [contig["name"] for contig in json_data["cgview"]["sequence"]["contigs"]]

    # Check that all the bed features are assigned to contigs that are in the JSON data.
    for bed_feature in bed_features:
        assert (
            bed_feature["contig"] in contigs
        ), "The contig {} listed in the BED file {} is not among the contigs in the genome being mapped.".format(
            bed_feature["contig"], bed_feature["legend"]
        )

    # Add the parsed BED features and tracks to the cgview map JSON data structure.
    json_data["cgview"]["features"] += bed_features
    json_data["cgview"]["tracks"] += bed_tracks
    # Write the cgview map JSON data structure to a new file.
    with open(output_file, "w") as f:
        json.dump(json_data, f, indent=4)


def parse_vcf_files(
    vcf_files: List[str],
) -> Tuple[List[Dict[str, Any]], List[Dict[str, Any]]]:
    """
    Parses VCF files.

    Parameters:
    vcf_files (list): A list of paths to VCF files.

    Returns:
    tuple: A tuple containing a list of parsed VCF features and a list of parsed VCF tracks.
    """
    vcf_features = []
    vcf_tracks = []
    for i, vcf_file in enumerate(vcf_files):
        num = i + 1

        # Define the VCF track.
        vcf_track = {
            "name": os.path.basename(vcf_file),
            "separateFeaturesBy": "none",
            "position": "both",
            "thicknessRatio": 1,
            "dataType": "feature",
            "dataMethod": "source",
            "dataKeys": f"vcf_{num}",
            "drawOrder": "score",
        }
        # Add the VCF track to the list of VCF tracks.
        vcf_tracks.append(vcf_track)

        # Parse the VCF file.
        with open(vcf_file) as f:
            vcf_results = f.readlines()
        # Parse the VCF file.
        for vcf_result_line in vcf_results:
            # Parse the VCF result line.
            vcf_result = vcf_result_line.strip().split()

            # Skip the header line(s), if present.
            if vcf_result[0].startswith("#") or vcf_result[0].startswith("\n"):
                continue

            assert (
                len(vcf_result) >= 5
            ), f"VCF file {vcf_file} is not in the correct format."

            name = ""
            if len(vcf_result) >= 8:
                name = vcf_result[7][20:]

            # Define the VCF feature.
            vcf_feature = {
                "name": name,
                "type": "vcf",
                "start": int(vcf_result[1]),
                "stop": int(vcf_result[1])
                + len(vcf_result[3])
                - 1,  # Assumes that the genome being mapped is the reference genome.
                "strand": 1,
                "source": f"vcf_{num}",
                "contig": vcf_result[0],
                "legend": os.path.basename(vcf_file),
                "tags": [],
                "meta": {
                    "ref": vcf_result[3],
                    "alt": vcf_result[4],
                },
            }

            # Add the VCF feature to the list of VCF features.
            vcf_features.append(vcf_feature)

    # Remove VCF features that are completely covered by another VCF feature
    # with an equal or higher score.
    vcf_features = [
        vcf_feature
        for vcf_feature, is_not_covered in zip(
            vcf_features,
            remove_covered_features(
                get_feature_locations_and_scores_from_vcf_features(vcf_features)
            ),
        )
        if is_not_covered
    ]

    assert (
        len(vcf_features) > 0 and len(vcf_tracks) > 0
    ), f"No VCF features or tracks were obtained from input file {vcf_file}."
    return (vcf_features, vcf_tracks)


def get_feature_locations_and_scores_from_vcf_features(
    vcf_features: List[Dict[str, Any]]
) -> List[Tuple[int, int, int]]:
    """
    Gets feature locations and scores from VCF features.

    Parameters:
    vcf_features (list): A list of parsed VCF features.

    Returns:
    list: A list of tuples containing feature locations and scores.
    """
    feature_locations_and_scores = []
    for vcf_feature in vcf_features:
        feature_locations_and_scores.append(
            (
                vcf_feature["start"],
                vcf_feature["stop"],
                1,
            )
        )
    return feature_locations_and_scores


def add_vcf_features_and_tracks(
    vcf_files: List[str], json_file: str, output_file: str
) -> None:
    """
    Parses VCF files, adds the parsed VCF features and tracks to the cgview map JSON data structure, and writes the cgview map JSON data structure to a new file.

    Parameters:
    vcf_files (list): A list of paths to VCF files.
    json_file (str): The path to a cgview map JSON file.
    output_file (str): The path to the output file.

    Returns:
    None
    """
    # Parse the VCF file.
    vcf_features, vcf_tracks = parse_vcf_files(vcf_files)
    # Read the cgview map JSON file.
    with open(json_file) as f:
        json_data = json.load(f)

    # Get list of contigs from JSON data.
    contigs = [contig["name"] for contig in json_data["cgview"]["sequence"]["contigs"]]

    # Check that all the vcf features are assigned to contigs that are in the JSON data.
    for vcf_feature in vcf_features:
        assert (
            vcf_feature["contig"] in contigs
        ), "The contig {} listed in the VCF file {} is not among the contigs in the genome being mapped.".format(
            vcf_feature["contig"], vcf_feature["legend"]
        )

    # Add the parsed VCF features and tracks to the cgview map JSON data structure.
    json_data["cgview"]["features"] += vcf_features
    json_data["cgview"]["tracks"] += vcf_tracks
    # Write the cgview map JSON data structure to a new file.
    with open(output_file, "w") as f:
        json.dump(json_data, f, indent=4)


# Use gffutils to parse GFF files.
def parse_gff_files(
    gff_files: List[str],
) -> Tuple[List[Dict[str, Any]], List[Dict[str, Any]]]:
    """
    Parses GFF files.

    Parameters:
    gff_files (list): A list of paths to GFF files.

    Returns:
    tuple: A tuple containing a list of parsed GFF features and a list of parsed GFF tracks.
    """
    gff_features = []
    gff_tracks = []
    for i, gff_file in enumerate(gff_files):
        num = i + 1

        # Define the GFF track.
        gff_track = {
            "name": os.path.basename(gff_file),
            "separateFeaturesBy": "none",
            "position": "both",
            "thicknessRatio": 1,
            "dataType": "feature",
            "dataMethod": "source",
            "dataKeys": f"gff_{num}",
            "drawOrder": "score",
        }
        # Add the GFF track to the list of GFF tracks.
        gff_tracks.append(gff_track)

        # Parse the GFF file.
        db = gffutils.create_db(
            gff_file,
            ":memory:",
            force=True,
            keep_order=True,
            merge_strategy="merge",
            sort_attribute_values=True,
        )

        # Parse the GFF file.
        for gff_result in db.all_features():
            # Skip the header line(s), if present.
            if gff_result.featuretype == "region":
                continue

            # Define the GFF feature.
            gff_feature = {
                "name": gff_result.id,
                "type": "gff",
                "start": gff_result.start,
                "stop": gff_result.stop,
                "strand": gff_result.strand,
                "source": f"gff_{num}",
                "contig": gff_result.seqid,
                "legend": os.path.basename(gff_file),
                "tags": [],
                "meta": {
                    "score": gff_result.score,
                },
            }
            # Add the GFF feature to the list of GFF features.
            gff_features.append(gff_feature)

    # Remove GFF features that are completely covered by another GFF feature
    # with an equal or higher score.
    gff_features = [
        gff_feature
        for gff_feature, is_not_covered in zip(
            gff_features,
            remove_covered_features(
                get_feature_locations_and_scores_from_gff_features(gff_features)
            ),
        )
        if is_not_covered
    ]

    assert (
        len(gff_features) > 0 and len(gff_tracks) > 0
    ), f"No GFF features or tracks were obtained from input file {gff_file}."
    return (gff_features, gff_tracks)


def get_feature_locations_and_scores_from_gff_features(
    gff_features: List[Dict[str, Any]]
) -> List[Tuple[int, int, int]]:
    """
    Gets feature locations and scores from GFF features.

    Parameters:
    gff_features (list): A list of parsed GFF features.

    Returns:
    list: A list of tuples containing feature locations and scores.
    """
    feature_locations_and_scores = []
    for gff_feature in gff_features:
        feature_locations_and_scores.append(
            (
                gff_feature["start"],
                gff_feature["stop"],
                gff_feature["meta"]["score"],
            )
        )
    return feature_locations_and_scores


def add_gff_features_and_tracks(
    gff_files: List[str], json_file: str, output_file: str
) -> None:
    """
    Parses GFF files, adds the parsed GFF features and tracks to the cgview map JSON data structure, and writes the cgview map JSON data structure to a new file.

    Parameters:
    gff_files (list): A list of paths to GFF files.
    json_file (str): The path to a cgview map JSON file.
    output_file (str): The path to the output file.

    Returns:
    None
    """
    # Parse the GFF file.
    gff_features, gff_tracks = parse_gff_files(gff_files)
    # Read the cgview map JSON file.
    with open(json_file) as f:
        json_data = json.load(f)

    # Get list of contigs from JSON data.
    contigs = [contig["name"] for contig in json_data["cgview"]["sequence"]["contigs"]]

    # Check that all the gff features are assigned to contigs that are in the JSON data.
    for gff_feature in gff_features:
        assert (
            gff_feature["contig"] in contigs
        ), "The contig {} listed in the GFF file {} is not among the contigs in the genome being mapped.".format(
            gff_feature["contig"], gff_feature["legend"]
        )

    # Add the parsed GFF features and tracks to the cgview map JSON data structure.
    json_data["cgview"]["features"] += gff_features
    json_data["cgview"]["tracks"] += gff_tracks
    # Write the cgview map JSON data structure to a new file.
    with open(output_file, "w") as f:
        json.dump(json_data, f, indent=4)
